// arm.c: 3DOF Arm Manipulator Program by Kosei Demura                                 
// copyright (c)2006  Kosei Demura

#include <ode/ode.h>                  // ODE
#include <drawstuff/drawstuff.h> // ODE̕`惉Cu
#define NUM 4                            // N(y܂)
dWorldID    world;                        // ͊wvZE
dBodyID     link[NUM];                 // N
dJointID      joint[NUM];              // ֐
static double THETA[NUM] = { 0.0, 0.0, 0.0, 0.0}; // ֐ߖڕWp
static double l[NUM]  = { 0.10, 0.90, 1.00, 1.00};  // N
static double r[NUM]  = { 0.20, 0.04, 0.04, 0.04};  // Na

void control() {
  static int step = 0;     // V~[ṼXebv
  double k1 =  10.0,  fMax  = 100.0;     // k1:QC
  printf("\r%6d:",step++);
  for (int j = 1; j < NUM; j++) {
    double tmpAngle = dJointGetHingeAngle(joint[j]); // ݊p
    double z = THETA[j] - tmpAngle;                        // z: c
    dJointSetHingeParam(joint[j],  dParamVel,  k1*z);
    dJointSetHingeParam(joint[j], dParamFMax, fMax);
    printf("THETA%d= %5.1f, ",j,tmpAngle/(M_PI/180.0));
  }
}
 
// `APȈ
void start() {             
  float xyz[3] = {  3.04, 1.28, 0.76};    // _
  float hpr[3] = { -160.0, 4.50, 0.00}; // (heading, pitch, roll)
  dsSetViewpoint(xyz,hpr);             // _Ǝ̐ݒ@@@
}

// L[֐
void command(int cmd) { //
  switch (cmd) {
  case 'j':  THETA[1] += 0.05; break;  // jL[
  case 'f':  THETA[1] -= 0.05; break;
  case 'k':  THETA[2] += 0.05; break;
  case 'd':  THETA[2] -= 0.05; break;
  case 'l':  THETA[3] += 0.05; break;
  case 's':  THETA[3] -= 0.05; break;
  }
  if (THETA[1] <   - M_PI)  THETA[1] =      - M_PI;
  if (THETA[1] >      M_PI)  THETA[1] =       M_PI;
  if (THETA[2] <  -2*M_PI/3)  THETA[2] =  - 2*M_PI/3;
  if (THETA[2] >   2*M_PI/3)  THETA[2] =   2*M_PI/3;
  if (THETA[3] <  -2*M_PI/3)  THETA[3] =  - 2*M_PI/3;
  if (THETA[3] >   2*M_PI/3)  THETA[3] =   2*M_PI/3;
}

// V~[V[v
void simLoop(int pause) {
  control();
  dWorldStep(world, 0.02);
  // {bg̕`                                                                    
  dsSetColor(1.0,1.0,1.0); // F̐ݒ
  for (int i = 0; i < NUM; i++ )
    dsDrawCappedCylinderD(dBodyGetPosition(link[i]),
          dBodyGetRotation(link[i]), l[i], r[i]);
} 
                                                  
int main(int argc, char *argv[]) {
  dsFunctions fn; // `֐
  dMass mass;  // ʃp[^
  double x[NUM] = {0.00}, y[NUM] = {0.00};              // dSʒu
  double z[NUM]         = { 0.05, 0.50, 1.50, 2.55};
  double m[NUM] = {10.00, 2.00, 2.00, 2.00};           // 
  double anchor_x[NUM]  = {0.00}, anchor_y[NUM] = {0.00};
  double anchor_z[NUM] = { 0.00, 0.10, 1.00, 2.00}  ;//]S@
  double axis_x[NUM]  = { 0.00, 0.00, 0.00, 0.00};     //]
  double axis_y[NUM]  = { 0.00, 0.00, 1.00, 1.00};
  double axis_z[NUM]  = { 1.00, 1.00, 0.00, 0.00};
  fn.version = DS_VERSION;  fn.start   = &start;
  fn.step   = &simLoop;      fn.command = &command;
  fn.path_to_textures = "../../drawstuff/textures";

  world = dWorldCreate();  // ͊wvZE̐
  dWorldSetGravity(world, 0, 0, -9.8);     // d͐ݒ
  
  for (int i = 0; i < NUM; i++) {                     // N̐Ɛݒ
    link[i] = dBodyCreate(world);                // N̐
    dBodySetPosition(link[i], x[i], y[i], z[i]); // ʒu̐ݒ
    dMassSetZero(&mass);                        // ʃp[^̏
    dMassSetCappedCylinderTotal(&mass,m[i],3,r[i],l[i]);  // N̎ʌvZ
    dBodySetMass(link[i], &mass);            // ʂ̐ݒ
  }
// ֐߂̐Ɛݒ
  joint[0] = dJointCreateFixed(world, 0); // Œ֐
  dJointAttach(joint[0], link[0], 0);          // ֐߂̎t
  dJointSetFixed(joint[0]);                      // ֐߂̐ݒ
  for (int j = 1; j < NUM; j++) {
    joint[j] = dJointCreateHinge(world, 0); // qW֐ߐ
    dJointAttach(joint[j], link[j-1], link[j]); // ֐߂̎t
    dJointSetHingeAnchor(joint[j], anchor_x[j], anchor_y[j],anchor_z[j]); //֐ߒS̐ݒ
    dJointSetHingeAxis(joint[j], axis_x[j], axis_y[j], axis_z[j]); // ֐߉]̐ݒ
  }
  dsSimulationLoop(argc, argv, 640, 570, &fn);
  return 0;
}
